@@ -32,6 +32,10 @@ enum {
 	FSLOC_STUB
 };
 
+#ifndef EXP_LOCKFILE
+#define EXP_LOCKFILE "/var/lib/nfs/export-lock"
+#endif
+
 typedef struct mclient {
 	struct mclient *	m_next;
 	char *			m_hostname;
@@ -137,6 +141,7 @@ nfs_export *			export_find(const struct addrinfo *ai,
 nfs_export *			export_allowed(const struct addrinfo *ai,
 						const char *path);
 nfs_export *			export_create(struct exportent *, int canonical);
+void				exportent_release(struct exportent *);
 void				export_freeall(void);
 int				export_export(nfs_export *);
 int				export_unexport(nfs_export *);
@@ -172,8 +177,6 @@ struct export_features {
 };
 
 struct export_features *get_export_features(void);
-
-/* Record export error.  */
-extern int export_errno;
+void fix_pseudoflavor_flags(struct exportent *ep);
 
 #endif /* EXPORTFS_H */
